# profile.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject
from gi.repository.GObject import ParamFlags
from pathlib import Path

from addwater.apps.firefox.firefox_paths import FirefoxPack


class Profile(GObject.Object):
    __gtype_name__ = "WaterProfile"

    name = GObject.Property(
        type=str, flags=(ParamFlags.CONSTRUCT_ONLY | ParamFlags.READWRITE)
    )

    # This is called "Path" by Firefox.
    id = GObject.Property(
        type=str, flags=(ParamFlags.CONSTRUCT_ONLY | ParamFlags.READWRITE)
    )
    favorite = GObject.Property(
        type=bool,
        default=False,
        flags=(ParamFlags.CONSTRUCT_ONLY | ParamFlags.READWRITE)
    )
    # TODO make this a GObject prop?
    path: Path
    package: FirefoxPack

    def __init__(self, name, id, path, favorite, package):
        super().__init__(name=name, id=id, favorite=favorite)
        self.path = Path(path)
        self.package = package

    # FIXME temporary prop until I redo ProfileSelector to use an icon instead
    @GObject.Property(type=str)
    def display_name(self) -> str:
        out = self.name
        if self.favorite:
            out += " (Preferred)"

        return out





# window.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later


import logging
from datetime import datetime, timezone
from os.path import exists, join

from addwater.page import AddWaterPage
from gi.repository import Adw, Gio, Gtk

from addwater import info

from .preferences import AddWaterPreferences
from .utils import paths

log = logging.getLogger(__name__)


@Gtk.Template(resource_path=info.PREFIX + "/gtk/window.ui")
class AddWaterWindow(Adw.ApplicationWindow):
    __gtype_name__ = "AddWaterWindow"

    main_menu = Gtk.Template.Child()
    # Use when only one page is available
    main_toolbar_view = Gtk.Template.Child()

    def __init__(self, backends: list, **kwargs):
        super().__init__(**kwargs)

        if info.PROFILE == "development":
            self.add_css_class("devel")

        self.set_size_request(360, 294)  # Minimum size of window Width x Height

        self.settings = Gio.Settings(schema_id=info.APP_ID)
        self.settings.bind(
            "window-height", self, "default-height", Gio.SettingsBindFlags.DEFAULT
        )
        self.settings.bind(
            "window-width", self, "default-width", Gio.SettingsBindFlags.DEFAULT
        )
        self.settings.bind(
            "window-maximized", self, "maximized", Gio.SettingsBindFlags.DEFAULT
        )
        self.create_action("preferences", self.on_preferences_action, ["<Ctrl>comma"])
        self.create_action("about", self.on_about_action)

        self.backends = backends
        self.create_pages(self.backends)

    # TODO just create a ViewStack and put the pages inside. Then hide the switcher widget
    #      if only one page is available. Makes it easy to switch to an error page
    def create_pages(self, app_backends: list):
        """Create and present app pages, and connect them to their respective app backend"""
        log.info("resetting gui pages")
        self.main_toolbar_view.set_content(None)
        self.pages = []

        for backend in app_backends:
            # Check data path for validity
            try:
                backend.get_package().get_profile_ini()
                page = AddWaterPage(backend=backend)
                log.debug("page created successfully")
            except FileNotFoundError as e:
                app_name = backend.get_app_name()
                log.critical(e)
                page = self.create_error_page(app_name)

            self.pages.append(page)

        if len(self.pages) == 1:
            self.main_toolbar_view.set_content(self.pages[0])
            log.debug("1 page available")
        else:
            log.error("multiple pages available")
            raise NotImplementedError

    def create_error_page(self, app_name: str):
        """Create basic error status page when the app faces a fatal error
        that must be communicated to the user.
        """
        # FIXME remove troubleshooting from this
        help_page_button = Adw.Clamp(
            hexpand=False,
            child=Gtk.Button(
                label=_("Open Help Page"),
                action_name="app.open-help-page",
                css_classes=["suggested-action", "pill"],
            ),
        )
        statuspage = Adw.StatusPage(
            # Translators: {} will be replaced with the app name ("Firefox" or "Thunderbird")
            title=_("{} Profile Data Not Found").format(app_name),
            # Translators: {} will be replaced with the app name ("Firefox" or "Thunderbird")
            description=_(
                "Please ensure {} is installed and Add Water has permission to access your profiles"
            ).format(app_name),
            child=help_page_button,
        )
        return statuspage

    def create_action(self, name: str, callback, shortcuts: list[str] = None):
        """Add a window action and shortcut.

        Args:
            name: the name of the action
            callback: the function to be called when the action is
              activated
            shortcuts: an optional list of accelerators
        """
        action = Gio.SimpleAction.new(name, None)
        action.connect("activate", callback)
        self.add_action(action)
        if shortcuts:
            app = self.get_application()
            app.set_accels_for_action(f"win.{name}", shortcuts)

    """Dialogs"""

    def on_preferences_action(self, *_args):
        """Callback for the app.preferences action."""
        pref = AddWaterPreferences()
        pref.present(self)

    def on_about_action(self, *_args):
        """Callback for the app.about action."""
        # Grab log info for debug info page
        now = datetime.now(timezone.utc).strftime("%Y-%m-%d")
        CURRENT_LOGFILE = join(paths.LOG_DIR, f"addwater_{now}.log")
        with open(file=CURRENT_LOGFILE, mode="r", encoding="utf-8") as f:
            db_info = f.read()

        # Setting up About dialog
        about = Adw.AboutDialog.new_from_appdata(
            (info.PREFIX + "/" + "dev.qwery.AddWater.metainfo"), info.VERSION
        )
        about.set_application_name("Add Water")
        about.set_application_icon(info.APP_ID)
        about.set_developer_name("qwery")
        about.set_version(info.VERSION)

        about.set_issue_url(info.ISSUE_TRACKER)
        about.set_website(info.WEBSITE)
        about.set_debug_info(db_info)
        about.set_debug_info_filename(f"addwater_{now}.log")

        about.set_developers(["Qwery"])
        # Translators: Replace this with "Your Name https://www.your-website.com" or "Your Name <your-email@example.com>"
        about.set_translator_credits(_("translator-credits"))
        about.set_copyright("© 2025 Qwery")
        about.set_license_type(Gtk.License.GPL_3_0)
        about.add_credit_section(
            # Translator: This is followed by a list of names
            name=_("Theme Created and Maintained by"),
            people=["Rafael Mardojai CM https://www.mardojai.com/"],
        )
        about.add_legal_section(
            "Other Wordmarks",
            "Firefox and Thunderbird are trademarks of the Mozilla Foundation in the U.S. and other countries.",
            Gtk.License.UNKNOWN,
            None,
        )

        about.present(self)



#TODO Make a class of Widgets that holds all this info as props, then define
#     these options in the UI Builder files instead

FIREFOX_OPTIONS = [
    {
        "group_name": _("Tab Position"),
        "options": [
            {
                "key": "hide-single-tab",
                "js_key": "hideSingleTab",
                "summary": _("Hide Single Tab"),
                "description": _("Hide Tab Bar when only one tab is open"),
                "tooltip": _(
                    "Consider placing the New Tab button near the Search Bar. Otherwise it may get lost when only one tab is open."
                ),
                "depends": None,
            },
            {
                "key": "normal-width-tabs",
                "js_key": "normalWidthTabs",
                "summary": _("Standard Tab Width"),
                "description": _("Use same tab width as standard Firefox"),
                "tooltip": "",
                "depends": None,
            },
            {
                "key": "swap-tab-close",
                "js_key": "swapTabClose",
                "summary": _("Swap Tab Close Button"),
                "description": _(
                    "Place Close Tab button on opposite side from window controls"
                ),
                "tooltip": "",
                "depends": None,
            },
            {
                "key": "bookmarks-toolbar-under-tabs",
                "js_key": "bookmarksToolbarUnderTabs",
                "summary": _("Bookmarks Toolbar Under Tabs"),
                "description": _("Place Bookmarks Bar underneath Tab Bar"),
                "tooltip": "",
                "depends": None,
            },
            {
                "key": "tabs-as-headerbar",
                "js_key": "tabsAsHeaderbar",
                "summary": _("Tabs as Headerbar"),
                "description": _("Place Tab Bar at top of window"),
                "tooltip": _(
                    'Enabling with "Hide Single Tab" will show a Title Bar when only one tab is open'
                ),
                "depends": None,
            },
            {
                "key": "tab-align-left",
                "js_key": "tabAlignLeft",
                "summary": _("Left-Aligned Titles"),
                "description": _("Align tab title to left"),
                "tooltip": "",
                "depends": None,
            },
        ],
    },
    {
        "group_name": _("Tab Behavior"),
        "options": [
            {
                "key": "active-tab-contrast",
                "js_key": "activeTabContrast",
                "summary": _("Active Tab Contrast"),
                "description": _("Add more visual contrast to the active tab"),
                "tooltip": "",
                "depends": None,
            },
            {
                "key": "close-only-selected-tabs",
                "js_key": "closeOnlySelectedTabs",
                "summary": _("Close Only Selected Tabs"),
                "description": _("Only show the Close Tab button on the active tab"),
                "tooltip": "",
                "depends": None,
            },
            {
                "key": "symbolic-tab-icons",
                "js_key": "symbolicTabIcons",
                "summary": _("Symbolic Tab Icons"),
                "description": _("Make all tab icons appear similar to symbolic icons"),
                "tooltip": "",
                "depends": None,
            },
            {
                "key": "all-tabs-button",
                "js_key": "allTabsButton",
                "summary": _("List of Tabs"),
                "description": _("Always show All Tabs button"),
                "tooltip": "",
                "depends": None,
            },
            {
                "key": "all-tabs-button-on-overflow",
                "js_key": "allTabsButtonOnOverflow",
                "summary": _("List of Tabs on Overflow"),
                "description": _("Only show All Tabs button when many tabs are open"),
                "tooltip": "",
                "depends": [("all-tabs-button", False)],
            },
        ],
    },
    {
        "group_name": "Style",
        "options": [
            {
                "key": "hide-webrtc-indicator",
                "js_key": "hideWebrtcIndicator",
                "summary": _("Hide Privacy Indicators"),
                "description": _("Hide microphone, screen share, and camera warnings"),
                "tooltip": _(
                    "GNOME will still show these privacy indicators in the Top Bar"
                ),
                "depends": None,
            },
            {
                "key": "oled-black",
                "js_key": "oledBlack",
                "summary": _("True Black Dark Theme"),
                "description": _(
                    "Use an OLED-friendly, deep black when dark theme is enabled"
                ),
                "tooltip": "",
                "depends": None,
            },
            {
                "key": "no-themed-icons",
                "js_key": "noThemedIcons",
                "summary": _("No Themed Icons"),
                "description": _(
                    "Use standard Firefox icons instead of Adwaita-styled icons"
                ),
                "tooltip": "",
                "depends": None,
            },
            {
                "key": "bookmarks-on-fullscreen",
                "js_key": "bookmarksOnFullscreen",
                "summary": _("Fullscreen Bookmarks"),
                "description": _("Show Bookmarks bar while in fullscreen"),
                "tooltip": "",
                "depends": None,
            },
        ],
    },
]



import logging
from typing import Callable, Optional

from gi.repository import Adw, Gio, GObject, Gtk

# TODO this whole thing is a mess and should just be statically
#      defined in a ui template
def create_option_group(
    group_schematic: dict[str, list[dict]],
    gui_switch_factory: Callable,
    settings,
):
    """PreferenceGroup factory for theme options.

    Args:
        group_schematic: A big dictionary list of all options to iterate through.
        gui_switch_factory: Function to use to create each switch for the group.
        settings: GSettings reader for the relevant app theme.

    Returns:
        AdwPreferenceGroup that's ready to use in the page.
    """

    group = Adw.PreferencesGroup(
        title=group_schematic["group_name"], margin_start=20, margin_end=20
    )

    for option in group_schematic["options"]:
        row = gui_switch_factory(
            title=option["summary"],
            subtitle=option["description"],
            extra_info=option["tooltip"],
        )

        row_switch = row.get_activatable_widget()
        settings.bind(
            option["key"], row_switch, "active", Gio.SettingsBindFlags.DEFAULT
        )

        # FIXME bind to the "active" property of the row it depends on.
        # requires holding a reference to that row which I don't atm

        # Handle dependencies on other options
        if option["depends"]:
            for prereq, b in option["depends"]:
                match b:
                    case True:
                        flag = Gio.SettingsBindFlags.DEFAULT
                    case False:
                        flag = Gio.SettingsBindFlags.INVERT_BOOLEAN
                settings.bind(prereq, row, "sensitive", flag)

        group.add(row)
    return group


def create_option_switch(title: str, subtitle: str, extra_info: Optional[str] = None):
    """Toggle row factory for theme options.

    Args:
        title: Title should be title-case and very short
        subtitle: One-sentence description to clarify what the option affects.
        extra_info: If needed, include disclaimers or extra important details
            inside an information button.

    Returns:
        AdwActionRow bound to its respective GSettings key.
    """
    row = Adw.ActionRow(title=title, subtitle=subtitle)
    # This styling was borrowed from GNOME settings > Mouse Acceleration option
    if extra_info:
        label = Gtk.Label(
            label=extra_info,
            margin_top=6,
            margin_bottom=6,
            margin_start=6,
            margin_end=6,
            max_width_chars=50,
            wrap=True,
        )
        info_popup = Gtk.Popover(autohide=True, child=label, hexpand=False)
        info_button = Gtk.MenuButton(
            has_frame=False,
            icon_name="info-outline-symbolic",
            valign="center",
            vexpand=False,
            tooltip_text=_("More Information"),
            popover=info_popup,
        )
        row.add_suffix(info_button)

    switch = Gtk.Switch(
        valign="center",
        vexpand=False,
    )
    row.add_suffix(switch)
    row.set_activatable_widget(switch)

    return row



# paths.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This module owns all app paths and must be used when referring to general paths like config or firefox paths

import os
from os.path import expanduser, join

from gi.repository import GLib

# APP PATHS
XDG_CACHE_DIR = GLib.get_user_cache_dir()
XDG_DATA_DIR = GLib.get_user_data_dir()
XDG_CONFIG_DIR = GLib.get_user_config_dir()

APP_CACHE = join(XDG_CACHE_DIR, "add-water")
APP_CONFIG = join(XDG_CONFIG_DIR, "add-water")
APP_DATA = join(XDG_DATA_DIR, "add-water")

DOWNLOAD_DIR = join(APP_CACHE, "downloads")
LOG_DIR = join(APP_CACHE, "logs")


def init_paths():
    paths = [APP_CACHE, DOWNLOAD_DIR, LOG_DIR]
    for each in paths:
        try:
            os.mkdir(path=each)
            print(f"{each} directory created.")
        except FileExistsError as err:
            print(f"{each} already exists. Skipped.")
        except FileNotFoundError as err:
            print("Couldn't find parent dir when initializing dirs ::", err)
            return

    print("All paths initialized.")
    print("-------------------------")



# install.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import shutil
from enum import Enum
from os import PathLike
from os.path import exists, join
from typing import Callable, Optional

from addwater.utils.paths import DOWNLOAD_DIR
from addwater.profile import Profile

log = logging.getLogger(__name__)


class InstallManager:
    """Provides API for all interactions that write or modify Firefox profiles.
    Make sure this is air-gapped and only accessed via strict pipelines to limit
    programmer errors that may lead to user data loss.

    Methods for installation, setting theme preferences, and uninstalling theme
    can be overridden by injecting your own functions. This leaves the door open
    for supporting different apps or themes in the future.

    args:
        installer: method that writes theme files into the app
        preference_handler: method to, if necessary, handle optional settings
            that may need to be written to a file somewhere.
        uninstaller: method to remove theme files from app. For now, this must
            handle removing those settings from pref_handler on its own.
    """

    def __init__(
        self,
        installer: Callable,
        preference_handler: Optional[Callable] = None,
        uninstaller: Optional[Callable] = None
    ):
        self._installer = installer

        self._preferences_handler = _set_theme_prefs
        if preference_handler:
            self._preferences_handler = preferences_handler

        self._uninstaller = _do_uninstall_theme
        if uninstaller:
            self._uninstaller = uninstaller

    """PUBLIC METHODS"""

    def combined_install(
        self,
        theme_path: PathLike,
        profile: Profile,
        options_results: Optional[dict[str, bool]] = None
    ) -> Enum:
        try:
            self._installer(
                profile=profile,
                theme_path=theme_path,
            )
            if options_results:
                self._preferences_handler(profile, options_results)
        except (InstallException, FileNotFoundError) as err:
            log.critical(err)
            return InstallStatus.FAILURE

        log.info("install complete")
        return InstallStatus.SUCCESS

    def uninstall(self, profile: Profile, folder_name):
        try:
            self._uninstaller(profile, folder_name)
        except InstallException as err:
            log.error(err)
            return InstallStatus.FAILURE

        return InstallStatus.SUCCESS


"""PRIVATE FUNCTIONS"""

"""Default install handlers. Can be overridden by injecting functions at construction."""


@staticmethod  # To avoid InstallManager passing self
def _set_theme_prefs(profile: Profile, options: dict[str, bool]) -> None:
    """Update user preferences in user.js according to GSettings.

    Args:
        profile: profile to install to
        options: the theme options list dicts
        gset_reader: Gio.Settings object preconfigured for the correct schema
                                        to read the values of the keys

    """
    log.info("Setting theme preferences in profile data...")

    user_js = join(profile.path, "user.js")
    # FIXME If the user.js file is gone, the other required prefs won't be set here
    # and thus the theme will not work properly
    try:
        with open(file=user_js, mode="r", encoding="utf-8") as file:
            lines = file.readlines()
    except FileNotFoundError:
        lines = []
    with open(file=user_js, mode="w", encoding="utf-8") as file:
        for key, value in options.items():
            pref_name = f"gnomeTheme.{key}"
            pref_value = str(value).lower()  # This MUST be lowercase
            full_line = f"""user_pref("{pref_name}", {pref_value});\n"""

            # TODO simplify this section
            found = False
            for i, line in enumerate(lines):
                # This is easier than a for-each
                if pref_name in line:
                    lines[i] = full_line
                    found = True
                    break
            if found is False:
                lines.append(full_line)
            log.debug(f"{pref_name} -> {pref_value}")

        file.writelines(lines)

    log.info("Done.")


@staticmethod  # To avoid InstallManager passing self
def _do_uninstall_theme(profile: Profile, theme_folder: str) -> None:
    log.info("Uninstalling theme from profile...")
    log.debug(f"Profile path: {profile.path}")

    # Delete theme folder
    chrome_path = join(profile.path, "chrome")
    try:
        theme_path = join(chrome_path, theme_folder)
        shutil.rmtree(theme_path)
    except FileNotFoundError:
        pass

    _remove_css_imports(chrome_path)

    # Set all user_prefs to false
    user_js = join(profile.path, "user.js")
    try:
        with open(file=user_js, mode="r", encoding="utf-8") as file:
            lines = file.readlines()
    except FileNotFoundError:
        log.info("Done.")
        return

    try:
        with open(file=user_js, mode="w", encoding="utf-8") as file:
            # TODO find a way to avoid using the index and just edit the line in the list directly
            for i, line in enumerate(lines):
                if "gnomeTheme" in line:
                    lines[i] = line.replace("true", "false")

            file.writelines(lines)
    except OSError as err:
        log.error(f"Resetting user.js prefs to false failed: {err}")
        raise InstallException("Uninstall failed")

    log.info("Done.")

def _remove_css_imports(chrome_path: PathLike) -> None:
    # TODO use this in the install process too
    # FIXME it only removes a ~third of duplicate imports at once. Why?
    css_files = ["userChrome.css", "userContent.css"]

    for each in css_files:
        p = join(chrome_path, each)
        try:
            with open(file=p, mode="r", encoding="utf-8") as file:
                log.debug(f"removing imports at {each}")
                lines = file.readlines()

            with open(file=p, mode="w", encoding="utf-8") as file:
                fgt_importline = f'@import "firefox-gnome-theme/{each}";'
                for line in lines:
                    if fgt_importline in line:
                        lines.remove(line)

                file.writelines(lines)
        except FileNotFoundError:
            log.error(f"couldn't find file {each}")
            pass

class InstallStatus(Enum):
    SUCCESS = 0
    FAILURE = 1


class InstallException(Exception):
    pass



# preferences.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

import gi

gi.require_version("Xdp", "1.0")

from addwater.backend import InterfaceMisuseError
from gi.repository import Adw, Gio, GObject, Gtk, Xdp

from addwater import info

log = logging.getLogger(__name__)


@Gtk.Template(resource_path=info.PREFIX + "/gtk/preferences.ui")
class AddWaterPreferences(Adw.PreferencesDialog):
    """Only used to change Add Water options.
    No theme options should be presented in this dialog. Include them in the
    future "Config" dialog section.
    """

    __gtype_name__ = "AddWaterPreferences"

    bg_notify_switch = Gtk.Template.Child()
    bg_update_switch = Gtk.Template.Child()

    def __init__(self):
        super().__init__()
        log.info("Preferences Window activated")
        self.settings_app = Gio.Settings(schema_id=info.APP_ID)
        self.portal = Xdp.Portal()

        # what did i mean by this?
        # TODO bind into Gsettings.
        self.settings_app.bind(
            "background-update",
            self.bg_update_switch,
            "enable-expansion",
            Gio.SettingsBindFlags.DEFAULT
        )
        self.settings_app.bind(
            "background-notifications",
            self.bg_notify_switch,
            "active",
            Gio.SettingsBindFlags.DEFAULT
        )

        self.bg_update_switch.connect(
            "activate",
            lambda *blah: self._do_background_request()
        )

    # FIXME make sure it doesn't run in the bg when called via the CLI
    def _do_background_request(self):
        """Request permission from portals to launch at login time"""
        bg_enabled = self.settings_app.get_boolean("background-update")
        if bg_enabled:
            flag = Xdp.BackgroundFlags.AUTOSTART
        else:
            flag = Xdp.BackgroundFlags.NONE

        self.portal.request_background(
            None,
            _("Checking for theme updates"),
            ["addwater", "--quick-update"],
            flag,
            None,
            None,
            None,
        )



# firefox_details.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from configparser import ConfigParser
from enum import Enum
from os.path import exists, join
from typing import Any, Callable, Optional
from pathlib import Path

from addwater.utils import paths
from gi.repository import Gio
from packaging.version import Version

from addwater import info
from addwater.profile import Profile
from .firefox_install import install_for_firefox
from .firefox_options import FIREFOX_OPTIONS
from .firefox_paths import FirefoxPack

log = logging.getLogger(__name__)

class FirefoxAppDetails:
    """
    name: proper, capitalized name of the app. i.e. 'Firefox' or 'Thunderbird
    options: theme options that the user can modify from the page GUI
    """

    # primary
    name: str = "Firefox"
    installed_version: Version

    # install
    installer: Callable = install_for_firefox
    options: list[dict[Any, Any]] = FIREFOX_OPTIONS
    package: FirefoxPack

    # online
    # TODO this should be defined as a const elsewhere
    THEME_URL: str = (
        "https://api.github.com/repos/rafaelmardojai/firefox-gnome-theme/releases"
    )

    # TODO just save the unzipped download in /tmp/ and then unzip into addwater/cache
    save_to = paths.DOWNLOAD_DIR
    app_folder = "firefox"
    theme_folder = "firefox-gnome-theme"
    full_path = join(save_to, app_folder, theme_folder)

    def __init__(self):
        self.settings = self.get_new_gsettings()

        version = Version(self.settings.get_string("installed-version"))
        self.set_installed_version(version)

        # If the path in GSettings is invalid, use the first available package
        current_path = self.settings.get_string("data-path")
        current_pack = FirefoxPack.new_from_path(Path(current_path))
        if not current_pack:
            available_packs = get_valid_packs()
            # TODO what should the app do if no packs are available?
            if not available_packs:
                log.critical("Could not find any valid data paths. App cannot function.")
                return
            self.set_package(available_packs[0])
            return

        self.set_package(current_pack)
        return

    """PUBLIC METHODS"""

    def reset_settings(self):
        """Resets all GSettings keys to their default"""
        log.info(f"Resetting all gsettings for {self.name}")
        self.settings.reset("theme-enabled")
        self.settings.reset("data-path")
        self.settings.reset("autofind-paths")
        self.settings.reset("installed-version")
        self.settings.reset("profile-selected")

        for group in self.options:
            for option in group["options"]:
                gset_key = option["key"]
                self.settings.reset(gset_key)
        log.info("done. gsettings reset")

    """Getters"""

    def get_new_gsettings(self):
        """Returns a ready-to-use Gsettings reader pre-configured for the relevant app theme."""
        log.debug(f"creating new Gsettings reader for {self.get_name()}")
        schema_id = info.APP_ID + "." + self.get_name()
        return Gio.Settings(schema_id=schema_id)

    def get_theme_folder_name(self):
        return self.theme_folder

    def get_download_path_info(self) -> tuple:
        """Returns a tuple of ([download cache path], [app folder], [theme files folder])

        Join these together to get the full path to the theme files for installation.
        """
        return (self.save_to, self.app_folder, self.theme_folder)

    def get_full_theme_path(self) -> str:
        return self.full_path

    def get_name(self) -> str:
        return self.name

    def get_package(self) -> FirefoxPack:
        return self.package

    def get_installer(self):
        return self.installer

    def get_installed_version(self):
        return self.installed_version

    def get_options(self):
        # TODO grab only the details the consumer would need.
        #      Multiple methods or add a flag?
        return self.options

    def get_profiles(self) -> list[Profile]:
        # TODO make this a set later to remove duplicates
        profiles = []

        for pack in FirefoxPack:
            profiles += find_profiles(pack)

        return profiles

    def get_info_url(self):
        return self.THEME_URL

    """Setters"""

    def set_package(self, package: FirefoxPack):
        # check if path is valid before setting
        _ini = package.get_profile_ini()

        self.settings.set_string("data-path", str(package.path))
        self.package = package


    def set_installed_version(self, new_version: Version) -> None:
        log.debug(f"Set installed version number to {new_version}")
        if not isinstance(new_version, Version):
            log.error("gave something that isn't a Version object. Fail")
            log.debug(f"input: {type(new_version)}")
            raise TypeError

        self.settings.set_string("installed-version", str(new_version))
        self.installed_version = new_version


def find_profiles(package: FirefoxPack) -> list[Profile]:
    cfg = ConfigParser()
    profiles = []

    try:
        profiles_ini = package.get_profile_ini()
    except FileNotFoundError:
        return []

    cfg.read(profiles_ini)
    for sect in filter(lambda sect: sect.startswith("Profile"), cfg.sections()):
        name = cfg[sect]["Name"]
        id = cfg[sect]["Path"]
        filepath = join(package.path, id)
        fav = False
        try:
            if cfg[sect]["Default"] == '1':
                fav = True
        except KeyError:
            pass

        profiles.append(Profile(name, id, filepath, fav, package))

    return profiles

def get_valid_packs() -> list[FirefoxPack]:
    found = []
    for pack in FirefoxPack:
        try:
            if (pack.get_profile_ini()):
                found.append(pack)
        except FileNotFoundError:
            pass

    return found



# page.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later


import logging
from typing import Callable, Optional

from addwater.gui.option_factory import (create_option_group,
                                         create_option_switch)
from gi.repository import Adw, Gio, GObject, Gtk
from packaging.version import Version

from addwater import info
from addwater.profile import Profile
from addwater.gui.profile_selector import ProfileSelector
from addwater.gui.pack_selector import PackSelector
from addwater.apps.firefox.firefox_paths import FirefoxPack

from .backend import InterfaceMisuseError

log = logging.getLogger(__name__)

@Gtk.Template(resource_path=info.PREFIX + "/gtk/addwater-page.ui")
class AddWaterPage(Adw.Bin):
    __gtype_name__ = "AddWaterPage"

    # Widget controls
    toast_overlay = Gtk.Template.Child()
    preferences_page = Gtk.Template.Child()
    change_confirm_bar = Gtk.Template.Child()

    general_pref_group = Gtk.Template.Child()

    enable_button = Gtk.Template.Child()
    profile_combobox = Gtk.Template.Child()
    package_combobox = Gtk.Template.Child()

    # TODO bind this to a prop that checks if the theme is installed
    theme_enabled = GObject.Property(
        type=bool,
        default=False,
        flags=(GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT)
    )

    app_name = GObject.Property(
        type=str,
        flags=(GObject.ParamFlags.READWRITE)
    )

    current_toast = None

    def __init__(self, backend):
        super().__init__()

        self.backend = backend

        # Stores changes as a transaction to apply all opts at once
        self.settings = self.backend.get_app_settings()
        self.settings.delay()

        self.init_gui(self.backend.get_app_options())
        self.init_actions()
        self.bind_settings()

        self.send_toast(_("Checking for updates..."), 10)
        self.request_update_status()

    """PUBLIC METHODS"""

    # TODO turn this into an async callback
    def request_update_status(self):
        update_status = self.backend.update_theme()
        match update_status:
            case update_status.UPDATED:
                if self.theme_enabled:
                    self.activate_action("apply-changes")

                v = self.backend.get_update_version()
                # Translators: {} will be replaced by a version number (example: "v126.5.65")
                msg = _("Updated theme to {}").format(str(v))
            case update_status.DISCONNECTED:
                msg = _("Failed to check for updates due to a network issue")
            case update_status.RATELIMITED:
                msg = _("Failed to check for updates. Please try again later.")
            case update_status.OTHER_ERROR:
                msg = _("Unknown error has occurred. Please report this to developer.")
            case update_status.NO_UPDATE:
                msg = None

        self.update_version_title()
        self.send_toast(msg)

    def on_apply_action(self, *_args):
        """Apply changes to GSettings and call the proper install or uninstall method"""
        log.debug("Applied changes")

        self.settings.apply()

        if self.theme_enabled:
            log.debug("GUI calling for install..")
            install_status = self.backend.begin_install(
                self.profile_combobox.get_selected_item(), True
            )
            toast_msg = _("Installed Theme. Restart Firefox to see changes.")
        else:
            log.debug("GUI calling for uninstall...")
            install_status = self.backend.remove_theme(
                self.profile_combobox.get_selected_item()
            )
            toast_msg = _("Removed Theme. Restart Firefox to see changes.")

        if install_status == install_status.FAILURE:
            toast_msg = _("Installation failed")

        self.send_toast(toast_msg, 3, 1)

    def on_discard_action(self):
        self.settings.revert()
        self.send_toast(_("Changes reverted"))

    def send_toast(
        self, msg: Optional[str] = None, timeout_seconds: int = 2, priority: int = 0
    ) -> None:
        """Convenience method to send an AdwToast quickly.

        Args:
            msg: Toast message as string. If none, it will still withdraw the
                 currently displayed toast.
            timeout_seconds: Default is 2 seconds.
            priority: default is Normal Priority (0). Use 1 to skip the toast queue
                and immediately present this toast.

        """
        if self.current_toast:
            self.current_toast.dismiss()

        if not msg:
            return

        self.current_toast = Adw.Toast(
            title=msg, timeout=timeout_seconds, priority=priority
        )

        self.toast_overlay.add_toast(self.current_toast)

        # Hack - AdwToasts don't timeout properly
        # details: https://gitlab.gnome.org/GNOME/libadwaita/-/issues/440
        self.enable_button.grab_focus()

    # TODO reduce all of this as much as possible using props and constructors
    def init_gui(self, options):
        """Create and bind all SwitchRows according to their respective GSettings keys

        Args:
            options: a json-style list of dictionaries which include all option groups
                    and options that the theme supports.
        """
        # Theme options
        for group_definition in options:
            group = create_option_group(
                group_schematic=group_definition,
                gui_switch_factory=create_option_switch,
                settings=self.settings,
            )
            self.bind_property(
                "theme-enabled",
                group, "sensitive",
                GObject.BindingFlags.DEFAULT | GObject.BindingFlags.SYNC_CREATE
            )
            self.preferences_page.add(group)

        pack = self.backend.get_package()
        self.package_combobox.setup_list(pack, self.backend)
        self.profile_combobox.setup_list(
            self.backend.get_profiles(),
            self.settings.get_string("profile-selected"),
            pack,
        )

    def bind_settings(self):
        # Primary Options
        self.settings.bind(
            "theme-enabled",
            self, "theme-enabled",
            Gio.SettingsBindFlags.DEFAULT
        )
        self.settings.bind("profile-selected", self.profile_combobox, "selected-profile-id", Gio.SettingsBindFlags.DEFAULT)
        self.settings.bind("autofind-paths", self.package_combobox, "autofind-paths", Gio.SettingsBindFlags.DEFAULT)

        # Change Confirmation bar
        self.settings.bind_property(
            "has-unapplied",
            self.change_confirm_bar,
            "revealed",
            GObject.BindingFlags.SYNC_CREATE,
        )

    def init_actions(self):
        action_group = Gio.SimpleActionGroup.new()

        apply_action = Gio.SimpleAction(name="apply-changes")
        apply_action.connect("activate", lambda *blah: self.on_apply_action())
        self.package_combobox.bind_property("valid-path", apply_action, "enabled", GObject.BindingFlags.SYNC_CREATE)
        action_group.add_action(apply_action)

        discard_action = Gio.SimpleAction(name="discard-changes")
        discard_action.connect("activate", lambda *blah: self.on_discard_action())
        action_group.add_action(discard_action)

        self.insert_action_group("water", action_group)

        # Combobox setup
        # TODO try to connect this in ui
        self.package_combobox.connect("package-changed", self.package_changed_cb)

    def package_changed_cb(self, pack_selector):
        if (pack := pack_selector.package):
            self.profile_combobox.update_package_filter(pack)

    def update_version_title(self):
        v = self.backend.get_update_version()
        version_str = parse_version_str(v)

        # Translators: {} will be replaced with a version number (example: v132) or a status message
        self.general_pref_group.set_title(_("Firefox GNOME Theme — {}").format(version_str))

def parse_version_str(version: Version):
    if version == Version("0.0.0"):
        return _("Not installed")

    return "v{}".format(str(version).removesuffix(".0"))




# main.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import os.path
import shutil
import sys
from datetime import datetime, timezone

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")

from addwater.apps.firefox.firefox_details import FirefoxAppDetails
from addwater.backend import BackendFactory
from gi.repository import Adw, Gio, GLib, Gtk

from addwater import info

from .utils import paths
from .utils.background import BackgroundUpdater
from .utils.logs import init_logs
from .window import AddWaterWindow

log = logging.getLogger(__name__)


class AddWaterApplication(Adw.Application):
    """The main application singleton class."""

    def __init__(self):
        super().__init__(
            application_id=info.APP_ID,
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE,
            resource_base_path=info.PREFIX,
        )
        self.create_action(
            "quit", lambda *_args: self.quit(), ["<primary>q", "<primary>w"]
        )
        self.create_action("open-help-page", self.on_help_action)
        self.create_action("reset-app", self.on_reset_app_action)

        paths.init_paths()
        init_logs()

        self.backends = self.construct_backends()

        self.add_main_option(
            "quick-update",
            ord("q"),
            GLib.OptionFlags.IN_MAIN,
            GLib.OptionArg.NONE,
            "Quickly update and install theme with the last-used settings",
            None,
        )

    def do_command_line(self, command_line):
        """Handles command line args and options if given, or starts the GUI
        window if none are provided.
        """
        options = command_line.get_options_dict()
        options = options.end().unpack()

        if options or info.FORCE_BG == "True":
            try:
                self.handle_background_update(options)
                return 0
            except CommandMisuseException as err:
                log.error(f"Use --help for proper usage notes: {err}")
                return 1

        self.activate()
        return 0

    def do_activate(self):
        """Called when the application is activated.

        We raise the application's main window, creating it if
        necessary.
        """

        # Create window with the logic it needs
        win = self.props.active_window
        if not win:
            win = AddWaterWindow(application=self, backends=self.backends)

        win.present()

    def handle_background_update(self, options):
        if info.FORCE_BG == "True":
            options = {"quick-update": True}

        # TODO add flag to reset app from CLI
        # TODO handle the option better and handle the error better
        if "quick-update" in options and options["quick-update"]:
            if not self.backends:
                log.error("Cannot find Firefox Profile Data")
                log.info(
                    "Please ensure Firefox is installed and Add Water has permission to access your profiles."
                )
                return

            background_updater = BackgroundUpdater(self.backends[0])
            background_updater.quick_update()

            notif = background_updater.get_status_notification()
            if notif:
                self.send_notification("addwater-bg-update-status", notif)
            return

        raise CommandMisuseException(f"Unknown options: {options}")

    def construct_backends(self):
        # TODO make this dynamic to find all available app details
        backends = []
        ff_app_detail = FirefoxAppDetails()
        backends.append(BackendFactory.new_from_appdetails(ff_app_detail))

        return backends

    def on_reset_app_action(self, *_args):
        log.warning("resetting the entire app...")

        settings = Gio.Settings(info.APP_ID)
        settings.reset("background-update")

        for each in self.backends:
            each.reset_app()

        try:
            shutil.rmtree(paths.DOWNLOAD_DIR)
        except FileNotFoundError:
            pass
        log.info("deleted download folder")

        log.info("app has been reset and will now exit")
        self.quit()

    def create_action(self, name, callback, shortcuts=None):
        """Add an application action.

        Args:
                name: the name of the action
                callback: the function to be called when the action is
                  activated
                shortcuts: an optional list of accelerators
        """
        action = Gio.SimpleAction.new(name, None)
        action.connect("activate", callback)
        self.add_action(action)
        if shortcuts:
            self.set_accels_for_action(f"app.{name}", shortcuts)

    def on_help_action(self, *_):
        log.info("help page action activated")
        weblaunch = Gtk.UriLauncher.new(info.TROUBLESHOOT_HELP)
        weblaunch.launch(None, None, None, None)


def main(version):
    """The application's entry point."""
    app = AddWaterApplication()
    return app.run(sys.argv)


class CommandMisuseException(Exception):
    pass



from os.path import expanduser, join, exists
from os import environ, PathLike
from enum import Enum
from pathlib import Path
from typing import Optional

config_dir = environ.get("HOST_XDG_CONFIG_HOME", expanduser("~/.config"))

# FIREFOX PATHS
# TODO replace fstring with join()
BASE_LEGACY = expanduser("~/.mozilla/firefox/")
BASE = expanduser(f"{config_dir}/mozilla/firefox/")
FLATPAK_LEGACY = expanduser("~/.var/app/org.mozilla.firefox/.mozilla/firefox/")
FLATPAK = expanduser("~/.var/app/org.mozilla.firefox/config/mozilla/firefox/")
SNAP_LEGACY = expanduser("~/snap/firefox/common/.mozilla/firefox/")
SNAP = expanduser("~/snap/firefox/common/.config/mozilla/firefox/")

LIBREWOLF_BASE_LEGACY = expanduser("~/.librewolf/")
LIBREWOLF_BASE = expanduser(f"{config_dir}/librewolf/librewolf")
LIBREWOLF_FLATPAK_LEGACY = expanduser(
    "~/.var/app/io.gitlab.librewolf-community/.librewolf/"
)
LIBREWOLF_FLATPAK = expanduser(
    "~/.var/app/io.gitlab.librewolf-community/config/librewolf/librewolf/"
)

FLOORP_BASE = expanduser("~/.floorp")
FLOORP_FLATPAK = expanduser("~/.var/app/one.ablaze.floorp/.floorp")

WATERFOX_BASE = expanduser("~/.waterfox")
WATERFOX_FLATPAK = expanduser("~/.var/app/net.waterfox.waterfox/.waterfox")

CACHY_BASE = expanduser("~/.cachy")

# TODO can this be a GObject so I can use it in ListModels?
#      or must I make a wrapper?
class FirefoxPack(Enum):
    # TODO rename these to just use the name instead of base later (Firefox / Firefox Flatpak / Librewolf / etc)
    BASE = ("Base", BASE)
    BASE_ALT = ("Base (Alt)", BASE_LEGACY)
    FLATPAK = ("Flatpak", FLATPAK)
    FLATPAK_ALT = ("Flatpak (Alt)", FLATPAK_LEGACY)
    SNAP = ("Snap", SNAP)
    SNAP_ALT = ("Snap (Alt)", SNAP_LEGACY)
    LIBREWOLF = ("Librewolf", LIBREWOLF_BASE)
    LIBREWOLF_ALT = ("Librewolf (Alt)", LIBREWOLF_BASE_LEGACY)
    LIBREWOLF_FLATPAK = ("Librewolf Flatpak", LIBREWOLF_FLATPAK)
    LIBREWOLF_FLATPAK_ALT = ("Librewolf Flatpak (Alt)", LIBREWOLF_FLATPAK_LEGACY)
    FLOORP = ("Floorp", FLOORP_BASE)
    FLOORP_FLATPAK = ("Floorp Flatpak", FLOORP_FLATPAK)
    WATERFOX = ("Waterfox", WATERFOX_BASE)
    WATERFOX_FLATPAK = ("Waterfox Flatpak", WATERFOX_FLATPAK)
    CACHY = ("CachyOS Browser", CACHY_BASE)

    def __init__(self, name: str, path: PathLike):
        self.pack_name = name
        self.path = Path(path)

    def get_profile_ini(self) -> Path:
        prof_ini = join(self.path, "profiles.ini")
        if not exists(prof_ini):
            raise FileNotFoundError(f"profiles.ini does not exist in {self.path}")

        return Path(prof_ini)


    @staticmethod
    def new_from_path(app_path: Path | PathLike) -> Optional[Enum]:
        app_path = Path(app_path)
        for pack in FirefoxPack:
            if app_path == pack.path:
                return pack

        return None

    @staticmethod
    def new_from_name(name: str) -> Optional[Enum]:
        for pack in FirefoxPack:
            if name == pack.pack_name:
                return pack

        return None




# background.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later


import logging
from enum import Enum
from typing import Any, Callable

from addwater import info
from addwater.backend import AddWaterBackend
from gi.repository import Gio

log = logging.getLogger(__name__)

# TODO add unit testing for this
# TODO clean up this whole thing
class BackgroundUpdater:
    """Simple class to update and install the theme without a GUI."""

    def __init__(self, backend: type[AddWaterBackend]):
        app_name = backend.get_app_name()
        log.debug(f"BackgroundUpdater created for {app_name}")
        self.backend = backend
        self.settings = self.backend.get_app_settings()

    def quick_update(self) -> None:
        update_status = self.backend.update_theme()
        match update_status:
            case update_status.UPDATED:
                status = self.silent_install()
            case update_status.NO_UPDATE:
                log.info("no update")
                status = SilentUpdateStatus.NO_UPDATE
            case update_status.DISCONNECTED:
                log.info("online failed, disconnected")
                status = SilentUpdateStatus.ONLINE_FAIL
            case update_status.RATELIMITED:
                log.info("online failed, rate limited")
                status = SilentUpdateStatus.ONLINE_FAIL
            case update_status.OTHER_ERROR:
                log.info("online failed. unknown error")
                status = SilentUpdateStatus.ONLINE_FAIL

        self.bg_status = status

        # TODO redo this to use the profile class and not rely on the id
    def silent_install(self):
        log.info("Update available. Silently installing")
        # TODO Move this check into backend.get_selected_profile()?

        selected_id = self.settings.get_string("profile-selected")
        profile_list = self.backend.get_profiles()
        profile = [p for p in profile_list if p.id == selected_id][0]

        # TODO if no profile, this should fail
        if not profile:
            profile = profile_list[0]

        install_status = self.backend.begin_install(profile, False)

        if install_status.SUCCESS:
            log.info("Silent install succeeded")
            return SilentUpdateStatus.UPDATED
        else:
            log.info("Silent install failed")
            return SilentUpdateStatus.INSTALL_FAIL

    def get_update_status(self) -> Enum:
        return self.bg_status

    def get_status_notification(self):
        log.debug("prepping a desktop notification for the bg update/install status")
        if not Gio.Settings(schema_id=info.APP_ID).get_boolean("background-notifications"):
            log.info("desktop notifications disabled")
            return None


        status = self.bg_status

        match status:
            case status.UPDATED:
                title = _("Theme updated")
                msg = _("Firefox GNOME Theme Updated")
            case status.ONLINE_FAIL:
                title = _("Theme Update Failed")
                msg = _("Could not check for theme updates due to a network issue")
            case status.INSTALL_FAIL:
                title = _("Theme Installation failed")
                msg = _("A theme update was downloaded but installation failed")
            case _:
                msg = None

        if msg:
            log.debug(f"notification text: {msg}")
            notif = Gio.Notification.new(title)
            notif.set_body(msg)
            notif.set_priority(Gio.NotificationPriority.LOW)
            return notif

        log.debug(f"nothing to report thus no notification sent")
        return None


class UpdaterException(Exception):
    pass


class SilentUpdateStatus(Enum):
    NO_UPDATE = 0
    UPDATED = 1
    ONLINE_FAIL = 2
    INSTALL_FAIL = 3



# logs.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

# TODO standardize all logs to make info logs more meaningful. Debug should include the tech details and info should only communicate the actual purpose and action that's being done and whether it was successful or not.

import logging
import os
import os.path
import sys
from datetime import datetime, timedelta, timezone

from addwater.info import PROFILE, VERSION
from gi.repository import Adw, Gtk

from . import paths


def init_logs():
    """Set up logging system"""
    LOG_DIR = paths.LOG_DIR
    try:
        now = datetime.now(timezone.utc).strftime("%Y-%m-%d")
        CURRENT_LOGFILE = os.path.join(LOG_DIR, f"addwater_{now}.log")

        file_handler = logging.FileHandler(CURRENT_LOGFILE)
        console_handler = logging.StreamHandler(sys.stdout)

        file_handler.setLevel(logging.DEBUG)
        if PROFILE == "development":
            console_handler.setLevel(logging.DEBUG)
        elif PROFILE == "default":
            console_handler.setLevel(logging.INFO)

        logging.basicConfig(
            handlers=[file_handler, console_handler],
            style="{",
            # TODO how to give a consistent spacing between levelname and name?
            format="[{levelname}]   {name} — {asctime} || {message}",
            datefmt="%H:%M",
            level=logging.DEBUG,
        )
    except Exception as err:
        print("Couldn't initialize log file: ", err)

    # Delete logs that are over two weeks old
    with os.scandir(path=LOG_DIR) as scan:
        oldest = ""
        for each in scan:
            time = datetime.strptime(
                each.name,
                "addwater_%Y-%m-%d.log",
            )
            time = time.replace(tzinfo=timezone.utc)
            difference = datetime.now(timezone.utc) - time
            if difference.days > 7:
                os.remove(os.path.join(LOG_DIR, each.name))

    info = f"""
	------------------------------------------------------------------------
	------------------------------------------------------------------------
	System Info:
	Add Water — An installer for the GNOME theme for Firefox and Thunderbird
	Version: {VERSION}
	Time (UTC): {datetime.now(timezone.utc)}
	GTK version: {Gtk.MAJOR_VERSION}.{Gtk.MINOR_VERSION}.{Gtk.MICRO_VERSION}
	ADW version: {Adw.MAJOR_VERSION}.{Adw.MINOR_VERSION}.{Adw.MICRO_VERSION}
	------------------------------------------------------------------------
	"""
    logging.debug(info)



# info.py.in
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later


APP_ID = 'dev.qwery.AddWater'
VERSION = '1.3'
PROFILE = 'default'
PREFIX = '/dev/qwery/AddWater'

# NOTE: Meson booleans are strings! Not native python bools. Could convert them here if that makes sense.
FORCE_BG = 'False'
MOCK_API = 'False'



# About Menu
ISSUE_TRACKER = 'https://github.com/largestgithubuseronearth/addwater/issues/'
WEBSITE = 'https://addwater.qwery.dev/'


# firefox_install.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later


import logging
import os
import shutil
from os import PathLike
from os.path import exists, join

from addwater.components.install import InstallException
from addwater.profile import Profile

log = logging.getLogger(__name__)


@staticmethod
def install_for_firefox(profile: Profile, theme_path: PathLike) -> None:
    """Install the Firefox theme. This method should be injected into the
    InstallManager at runtime. If it isn't obvious, this should not be reused for
    installing other app themes.

    Args:
        theme_path: path to the extracted theme folder. Likely inside `[app_path]/cache/add-water/downloads/`
        profile_path: path to the profile folder in which the theme will be installed.

    """
    # Check paths to ensure they exist
    log.info("Installing theme file for Firefox...")
    if not exists(profile.path):
        raise FileNotFoundError("Install failed. Profile path not found.")
    if not exists(theme_path):
        raise FileNotFoundError("Install failed. Theme files not found.")

    chrome_path = join(profile.path, "chrome")

    _copy_files(chrome_path, theme_path)
    _import_css(chrome_path)

    userjs_template = join(
        chrome_path, "firefox-gnome-theme", "configuration", "user.js"
    )
    _copy_userjs(profile.path, userjs_template)

    log.info("Firefox installation done.")


def _copy_files(chrome_path: str, theme_path: str):
    # Make chrome folder if it doesn't already exist
    log.debug("Copying theme files into profile path...")
    try:
        os.mkdir(chrome_path)
    except FileNotFoundError:
        log.critical("Profile path does not exist. Install canceled.")
        raise InstallException("Profile doesn't exist.")
    except FileExistsError:
        pass

    # Copy theme repo into chrome folder
    shutil.copytree(
        src=theme_path, dst=join(chrome_path, "firefox-gnome-theme"), dirs_exist_ok=True
    )
    log.debug("Done.")


def _import_css(chrome_path: str):
    log.debug("Adding CSS imports...")
    css_files = ["userChrome.css", "userContent.css"]

    for each in css_files:
        p = join(chrome_path, each)
        try:
            with open(file=p, mode="r", encoding="utf-8") as file:
                lines = file.readlines()
        except FileNotFoundError:
            log.debug(f"{each} does not exist. Creating it from scratch")
            lines = []

        with open(file=p, mode="w", encoding="utf-8") as file:
            # Remove old import lines
            for line in lines:
                if "firefox-gnome-theme" in line:
                    lines.remove(line)

            import_lines = []
            import_lines.append(f'@import "firefox-gnome-theme/{each}";\n')

            file.writelines(import_lines + lines)

        log.debug(f"Finished importing {each}")
    log.debug("Done.")


def _copy_userjs(profile_path: str, template_path: str) -> None:
    """Append content of user.js template to the existing user.js if the URL doesn't exist."""
    log.debug("Checking if URL already exists in user.js...")
    user_js = join(profile_path, "user.js")
    user_js_backup = join(profile_path, "user.js.bak")
    url_to_check = "https://github.com/rafaelmardojai/firefox-gnome-theme/"

    # Check if the (Required user.js content) already exists in user.js
    if exists(user_js):
        with open(user_js, "r", encoding="utf-8") as user_js_file:
            content = user_js_file.read()
            if url_to_check in content:
                log.debug("The specified URL already exists in user.js.")
                return True  # URL exists, exit the function

    # Backup user.js if it exists and backup does not exist
    if exists(user_js) and not exists(user_js_backup):
        log.debug("Backing up user's previous user.js file")
        shutil.copy(user_js, user_js_backup)

    # Append the content of the template user.js to the existing user.js
    try:
        with open(template_path, "r", encoding="utf-8") as template_file:
            template_content = template_file.read()

        with open(user_js, "a", encoding="utf-8") as user_js_file:
            user_js_file.write(template_content)
            log.debug("Appended user.js content successfully.")

    except Exception as e:
        log.critical(f"Failed to append to user.js: {e}")
        raise InstallException("Failed to append user.js")

    log.debug("Done.")



# online.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later


import json
import logging
import os
import shutil
import tarfile
from enum import Enum
from os.path import exists, join
from typing import Optional

import requests
from addwater.utils.paths import DOWNLOAD_DIR
from packaging.version import Version

from addwater import info

log = logging.getLogger(__name__)


# TODO make this handle requests asynchronously and use callbacks to report the status
# This would allow the app to launch instantly, even if internet speed is poor


class OnlineManager:
    """Handles everything to do with polling for theme updates and downloading
    release packages. It will also handle prepping those releases for the installer to use.

    Args:
        theme_url: URL to the theme's Github repo

    """

    update_version: Version
    theme_url: str

    def __init__(self, theme_url: str):
        log.debug("online manager is now alive")
        self.theme_url = theme_url

    """PUBLIC METHODS"""

    def get_updates_online(self, installed_version: Version, path_info: tuple) -> Enum:
        """Check for updates and if one is available, download and install it.

        Args:
            installed_version: Currently installed version to compare against update.
            path_info: Where to download the update files to.
        """
        log.info("Checking for updates...")
        self.update_version = installed_version

        try:
            update_info = self._get_release_info(self.theme_url)
        except NetworkException:
            return OnlineStatus.DISCONNECTED

        # Decide whether we need to update or not
        calls_left = update_info["ratelimit_remaining"]
        if self._is_ratelimit_exceeded(calls_left):
            log.warning("rate limiting self to avoid angering Github")
            return OnlineStatus.RATELIMITED

        files_downloaded = exists(join(path_info[0], path_info[1], path_info[2]))

        if (
            self._is_update_available(update_info["version"], installed_version)
            or not files_downloaded
        ):
            log.info("update available. getting it now...")
            self.update_version = update_info["version"]
            return self._begin_download(path_info, update_info["tarball_url"])

        log.info("no update available")
        return OnlineStatus.NO_UPDATE

    def get_update_version(self):
        return self.update_version

    """PRIVATE METHODS"""

    def _begin_download(self, path_info, tarball_url) -> Enum:
        # Update if necessary
        # TODO simplify to just pass the path info into get_release
        base_path = join(path_info[0], path_info[1])
        final_name = path_info[2]
        try:
            self._get_release(
                base_path=base_path, final_name=final_name, tarball_url=tarball_url
            )
            return OnlineStatus.UPDATED
        except NetworkException as err:
            return OnlineStatus.DISCONNECTED
        except ExtractionException as err:
            log.error(err)
            return OnlineStatus.OTHER_ERROR  # TODO handle this error better

    def _get_release(self, base_path: str, final_name: str, tarball_url: str) -> None:
        """Download and prep a theme release for installati

        Args:
            base_name: the naming convention for the download zipfile and extracted path
            final_path: naming convention for the resulting theme files. As of now,
                    this is '{app_name}-gnome-theme'. This means you can find the theme
                    folder at "{base-name}-extracted/{final_path}"
        """
        zipfile = f"{base_path}.tar.gz"
        extract_path = f"{base_path}"
        final_path = join(extract_path, final_name.lower())

        log.info(f"Getting release...")

        if not exists(zipfile) or not exists(extract_path):
            try:
                self._download_tarball(tarball_url, zipfile)
            except (requests.RequestException, requests.ConnectionError) as err:
                log.error(err)
                raise NetworkException(err)

        try:
            shutil.rmtree(final_path)
        except FileNotFoundError:
            pass

        try:
            self._extract_tarball(zipfile, extract_path)
        except (FileNotFoundError, tarfile.TarError) as err:
            raise ExtractionException("Theme files failed to extract")

        # rename inner folder
        self._rename_theme_folder(extract_path, final_name)

        log.info("Update files downloaded and ready to install.")

    @staticmethod
    def _download_tarball(dl_url: str, result: str) -> None:
        """Download file and write to a file

        Args:
            dl_url: url to the file to download
            result: path to write the contents of the downloaded file
        """
        if exists(result):
            log.debug("Already downloaded this release. Skipping download.")
            return

        headers = {
            "X-Github-Api-Version": "2022-11-28",
            "User-Agent": (info.APP_ID + "/" + info.VERSION),
            "Accept": "application/vnd.github.x-gzip+json",  # Note: I'm not sure if this is an acceptable way to do it'
        }
        response = requests.get(dl_url, headers=headers, timeout=10)

        with open(file=result, mode="wb") as file:
            file.write(response.content)

        log.debug("Successfully downloaded release from Github")

    @staticmethod
    def _extract_tarball(zipfile_path: str, result_path: str) -> None:
        """Extracts tar.gz files. It's important to know that this destroys the tar after the extraction is done.

        Args:
            zipfile_path: input tar.gz file to unzip
            result_path: resulting extracted directory
        """

        if not exists(zipfile_path):
            raise FileNotFoundError(
                "Zipfile does not exist to be extracted. It must have been lost since downloading it"
            )

        with tarfile.open(zipfile_path) as tf:
            tf.extractall(path=result_path, filter="data")
        log.debug(f"Successfully unzipped the file to {result_path}")

        os.remove(zipfile_path)
        log.debug(f"Deleted {zipfile_path}")

    @staticmethod
    def _rename_theme_folder(parent_dir: str, new_name: str) -> None:
        """Renames the inner folder of the release to be easy to find later."""
        final_path = join(parent_dir, new_name)
        if not exists(parent_dir):
            raise FileNotFoundError("Rename failed. Parent folder does not exist.")
        if exists(final_path):
            log.debug("Inner file is already properly named. Skipping rename step.")
            return

        with os.scandir(path=parent_dir) as scan:
            for each in scan:
                if each.name.startswith("rafaelmardojai"):
                    old_path = join(parent_dir, each.name)
                    os.rename(old_path, final_path)
        log.debug("Successfully renamed inner folder")

    @staticmethod
    def _get_release_info(gh_url: str) -> dict:
        """Poll Github url and check if a new release of the theme is available

        Args:
            gh_url: fully-qualified url to a github api releases endpoint
        Returns:
            release_info: dict including "version" as int, "ratelimit_remaining" as int, and "tarball_url" as str
        """
        # TODO If you can check Firefox version easily, check that before polling GH

        headers = {
            "X-Github-Api-Version": "2022-11-28",
            "User-Agent": (info.APP_ID + "/" + info.VERSION),
            "Accept": "application/vnd.github+json",
        }
        try:
            response = requests.get(gh_url, headers=headers, timeout=10)
        except requests.RequestException as err:
            # TODO use specific exceptions to handle being disconnected. It'll
            # 	be more helpful if the issue isn't just being offline but an API or
            # 	programmer error
            log.error(f"Could not connect to Github to grab release info: {err}")
            raise NetworkException(err)

        api_calls_left = int(response.headers["x-ratelimit-remaining"])
        try:
            latest_release = response.json()[0]
            version = Version(latest_release["tag_name"])
            tarball_url = latest_release["tarball_url"]
        except requests.JSONDecodeError as err:
            log.error(err)
            version = None
            tarball_url = None

        # TODO make this a tuple instead for easy unpacking
        release_info = {
            "version": version,
            "ratelimit_remaining": api_calls_left,
            "tarball_url": tarball_url,
        }
        return release_info

    @staticmethod
    def _is_ratelimit_exceeded(api_calls_left: int) -> bool:
        # TODO improve this?

        # Maybe set the time and api calls remaining in gsettings. There is
        # only a warning and no mechanism stopping user from continuing to spam.
        CHOSEN_LIMIT = 10
        log.debug(f"Remaining Github API calls for the next hour: {api_calls_left}")
        return bool(api_calls_left < CHOSEN_LIMIT)

    @staticmethod
    def _is_update_available(new: Version, current: Version) -> bool:
        if not isinstance(current, Version) or not isinstance(new, Version):
            raise TypeError(
                "checking updates requires both values to be Version objects"
            )

        return new > current


class OnlineStatus(Enum):
    NO_UPDATE = 0
    UPDATED = 1
    DISCONNECTED = 2
    RATELIMITED = 3
    OTHER_ERROR = 4


class NetworkException(Exception):
    pass


class OnlineManagerError(Exception):
    pass


class ExtractionException(Exception):
    pass



# profile_selector.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from typing import Callable, Optional

from gi.repository import GObject, Gtk, Adw, Gio
from packaging.version import Version

from addwater import info
from addwater.profile import Profile
from addwater.backend import InterfaceMisuseError
from addwater.apps.firefox.firefox_paths import FirefoxPack

log = logging.getLogger(__name__)

# TODO merge the package selector into this one combobox, and break them into
#      sections by package format

@Gtk.Template(resource_path=info.PREFIX + "/gtk/profile-selector.ui")
class ProfileSelector(Adw.ComboRow):
    __gtype_name__ = "AddWaterProfileSelector"

    profiles: Gio.ListStore = Gtk.Template.Child()
    sort_model: Gtk.SortListModel = Gtk.Template.Child()
    sorter: Gtk.CustomSorter = Gtk.Template.Child()

    filter_model: Gtk.FilterListModel = Gtk.Template.Child()
    filter: Gtk.CustomFilter = Gtk.Template.Child()

    # GSettings needs this to store profile id for now
    selected_profile_id = GObject.Property(type=str, flags=(GObject.ParamFlags.READWRITE))

    def __init__(self):
        super().__init__()

        self.sorter.set_sort_func(sort_profiles)

    # TODO move this into constructor
    def setup_list(self, profile_list, selected_profile_id, pack):
        self.profiles.splice(0, self.profiles.get_n_items(), profile_list)
        self.update_package_filter(pack)

        # TODO simplify this mess
        for i, profile in enumerate(self.filter_model):
            if profile.id == selected_profile_id:
                self.set_selected(i)
                return

        self.set_selected(0)

    def update_package_filter(self, pack):
        self.filter.set_filter_func(filter_profiles, pack)
        self.filter.changed(Gtk.FilterChange.DIFFERENT)

def sort_profiles(a: Profile, b: Profile, _data) -> int:
    if (res := b.favorite - a.favorite) != 0:
        return res

    return (a.name > b.name) - (b.name > a.name)

def filter_profiles(item: Profile, new_package: FirefoxPack):
    if not new_package:
        return True

    return item.package == new_package



# page.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later


import logging
from typing import Callable, Optional

from gi.repository import Adw, Gio, GObject, Gtk

from addwater import info
from addwater.backend import InterfaceMisuseError
from addwater.apps.firefox.firefox_paths import FirefoxPack

log = logging.getLogger(__name__)


@Gtk.Template(resource_path=info.PREFIX + "/gtk/pack-selector.ui")
class PackSelector(Adw.ComboRow):
    __gtype_name__ = "WaterPackSelector"

    # TODO backend shouldn't be here at all
    str_model: Gtk.StringList = Gtk.Template.Child()

    backend = None
    package: FirefoxPack
    inner_valid_path: bool
    autofind_paths: bool = GObject.Property(type=bool, default=True, flags=GObject.ParamFlags.READWRITE)

    def __init__(self):
        super().__init__()
        self.notify('selected-item')
        self.connect('notify::selected-item', lambda *args: self.set_package())


    # TODO this should be done in constructor; shouldn't need Page's help to do this
    def setup_list(self, pack, backend):
        self.backend = backend

        # FIXME HACK. the autofind setting is only bound after this method is called.
        # this would be fixed by not doing it this way to begin with
        self.autofind_paths = self.backend.get_app_settings().get_boolean('autofind-paths')

        # Don't remove the first item, Auto discover
        self.str_model.splice(1, len(self.str_model) - 1, [p.pack_name for p in FirefoxPack])

        if pack and not self.autofind_paths:
            self.package = pack
            # Offset 1 since Auto Discover is first
            i = list(FirefoxPack).index(pack) + 1
            self.set_selected(i)

    def set_package(self):
        selected_index = self.get_selected()
        AUTO = 0

        if selected_index == AUTO:
            self.autofind_paths = True
            log.info("Autofind paths enabled")
            self.valid_path = True
            self.emit("package-changed")
            return

        self.autofind_paths = False
        log.info("Autofind paths disabled")

        packstr = self.get_selected_item().get_string()
        new_pack = FirefoxPack.new_from_name(packstr)

        # TODO redo this so i don't call backend. either this widget should be
        #      responsible for packages altogether OR it should verify packages
        #      itself and Page will commit the change only if it's valid.
        if new_pack:
            try:
                self.backend.set_package(new_pack)
            except InterfaceMisuseError as err:
                log.error(err)
                self.valid_path = False
            else:
                self.valid_path = True
                self.package = new_pack
                self.emit("package-changed")

        return

    @GObject.Property(type=bool, default=False)
    def valid_path(self) -> bool:
        return self.inner_valid_path

    @valid_path.setter
    def set_valid_path(self, is_valid: bool):
        if is_valid:
            self.remove_css_class("error")
        else:
            self.add_css_class("error")

        self.inner_valid_path = is_valid

    @GObject.Signal(name="package-changed")
    def package_changed(self):
        pass



# backend.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from enum import Enum
from os.path import exists, join
from typing import Any, Callable, Optional
from pathlib import Path

from addwater.profile import Profile
from addwater.components.install import (InstallManager, InstallStatus)
from addwater.components.online import OnlineManager
from addwater.utils.mocks import mock_online
from addwater.utils.paths import DOWNLOAD_DIR
from addwater.apps.firefox.firefox_paths import FirefoxPack
from packaging.version import Version

from addwater import info

log = logging.getLogger(__name__)

# TODO remove this and let the page use the individual components directly.
#      This requires everything to be bound correctly as GObjects

class AddWaterBackend:
    """Interface to perform high-level app actions like installing, getting
    profile information, data paths, and so on. Only this class should ever
    interact with its own provider components directly.

    This class can live without a GUI frontend to allow for background updating.

    Args:
        app_details: object to track the state of important app details
            such as the app's name, available profiles, data paths, and what
            theme options are available. Will be passed into other managers
            for convenience.
        install_manager: object that manages processes related to installing,
            uninstalling, and setting user's theme preferences in
            configuration files.
        online_manager: object that makes internet requests, processes http responses,
            handles network errors, downloads theme releases, and preps
            those releases to be installed.
    """

    def __init__(
        self,
        app_details,
        install_manager: type[InstallManager],
        online_manager: type[OnlineManager],
    ):
        self.app_details = app_details
        self.install_manager = install_manager
        self.online_manager = online_manager

        log.info(f"Backend created for {self.get_app_name()}")

    """PUBLIC METHODS

	For these methods, it's often impossible to avoid modifying object states. So it's
	important for these methods to be readable and concise, and modify as little
	as possible.
	"""

    """Install actions"""

    def begin_install(self, profile: Profile, full_install=False) -> Enum:
        log.info("beginning installation...")
        if not profile:
            return InstallStatus.FAILURE

        # FIXME this path stuff sucks
        result = self.app_details.get_download_path_info()
        theme_path = join(result[0], result[1], result[2])

        # Prep options for pref handler if full install
        if full_install:
            options_request = {}
            gset = self.app_details.get_new_gsettings()
            options = self.app_details.get_options()

            # TODO redo the options entirely into structs
            for group in options:
                for option in group["options"]:
                    js_key = option["js_key"]
                    # TODO shouldn't need to check gsettings at this point.
                    #      associate the value to the option directly.
                    gset_value = gset.get_boolean(option["key"])

                    options_request[js_key] = gset_value
        else:
            options_request = None

        status = self.install_manager.combined_install(
            theme_path=theme_path,
            profile=profile,
            options_results=options_request,
        )

        log.info("install process completed")
        return status

    def remove_theme(self, profile: Profile) -> Enum:
        folder_name = self.app_details.get_theme_folder_name()

        return self.install_manager.uninstall(profile, folder_name)

    """Online Actions"""

    def update_theme(self) -> Enum:
        status = self.online_manager.get_updates_online(
            self.get_installed_version(),
            self.app_details.get_download_path_info(),
        )

        self.set_installed_version(self.get_update_version())

        return status

    """Info Getters"""

    def get_app_name(self) -> str:
        return self.app_details.get_name()

    def get_app_settings(self):
        return self.app_details.get_new_gsettings()

    def get_app_options(self) -> list[dict[str, Any]]:
        return self.app_details.get_options()

    def get_package(self) -> FirefoxPack:
        return self.app_details.get_package()

    def get_installed_version(self) -> Version:
        return self.app_details.get_installed_version()

    def get_update_version(self) -> Version:
        return self.online_manager.get_update_version()

    def get_profiles(self) -> list[Profile]:
        return self.app_details.get_profiles()

    """Info Setters"""

    def set_package(self, pack: FirefoxPack):
        try:
            self.app_details.set_package(pack)
        except FileNotFoundError as err:
            raise InterfaceMisuseError(f"Invalid path: {err}")

    def set_installed_version(self, new_version: Version) -> None:
        if not isinstance(new_version, Version):
            raise InterfaceMisuseError(
                f"method takes Version object; gave {type(new_version)}"
            )
        self.app_details.set_installed_version(new_version)

    def reset_app(self):
        self._uninstall_all_profiles()
        self.app_details.reset_settings()
        log.info(f"done. Add Water has been reset to default state")

    """PRIVATE METHODS"""

    # TODO redo to use Profile class
    def _uninstall_all_profiles(self):
        pass

# TODO probably should just remove these?
class InterfaceMisuseError(Exception):
    pass


class FatalInterfaceError(Exception):
    pass


class BackendFactory:

    @staticmethod
    def new_from_appdetails(app_details):
        install_method = app_details.get_installer()
        install_manager = InstallManager(
            installer=install_method,
        )

        if info.MOCK_API == "True":
            online_manager = mock_online.MockOnlineManager()
        else:
            theme_url = app_details.get_info_url()
            online_manager = OnlineManager(
                theme_url=theme_url,
            )

        firefox_backend = AddWaterBackend(
            app_details=app_details,
            install_manager=install_manager,
            online_manager=online_manager,
        )
        return firefox_backend



# download.py
#
# Copyright 2025 Qwery
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from enum import Enum
from typing import Optional

from gi.repository import Gio

from addwater import info

from packaging.version import Version

log = logging.getLogger(__name__)


class MockOnlineManager:
    """PUBLIC METHODS"""

    def __init__(self):
        log.warning("Mock online manager created!!!")
        self.online_status = OnlineStatus.UPDATED
        schema_id = info.APP_ID + ".Firefox"
        self.settings = Gio.Settings(schema_id=schema_id)

        self.update_version = Version(self.settings.get_string("installed-version"))

    def get_updates_online(self, *args, **kwargs) -> Enum:
        log.debug(f"returning fake status code of {self.online_status}")
        return self.online_status

    def get_release(self, base_name: str, final_name: str, tarball_url: str):
        pass

    def get_update_version(self):
        return self.update_version


class OnlineStatus(Enum):
    NO_UPDATE = 0
    UPDATED = 1
    DISCONNECTED = 2
    RATELIMITED = 3


class NetworkException(Exception):
    pass


class OnlineManagerError(Exception):
    pass


class ExtractionException(Exception):
    pass
